@@ -1,11 +1,13 @@
 #include "config.h"
 #include "aworkspaces.h"
 #include "workspaces.h"
-#include "prefs.h"
+#include "default.h"
+#include "yprefs.h"
 #include "yxapp.h"
 #include "wmframe.h"
+#include "wmmgr.h"
 #include "udir.h"
-#include "ypaths.h"
+#include "wmapp.h"
 #include "wpixmaps.h"
 #include "intl.h"
 #include <math.h>
@@ -18,21 +20,31 @@ YColorName WorkspaceButton::activeButtonBg(&clrWorkspaceActiveButton);
 YColorName WorkspaceButton::activeBackupBg(&clrActiveButton);
 YColorName WorkspaceButton::activeButtonFg(&clrWorkspaceActiveButtonText);
 
-ref<YFont> WorkspaceButton::normalButtonFont;
-ref<YFont> WorkspaceButton::activeButtonFont;
+YFont WorkspaceButton::normalButtonFont;
+YFont WorkspaceButton::activeButtonFont;
+
+ref<YImage> WorkspaceButton::normalGradient;
+ref<YImage> WorkspaceButton::activeGradient;
+
+void WorkspaceButton::freeFonts() {
+    normalButtonFont = null;
+    activeButtonFont = null;
+    normalGradient = null;
+    activeGradient = null;
+}
 
 WorkspaceButton::WorkspaceButton(int ws, YWindow *parent, WorkspaceDragger* d):
     super(parent, YAction()),
     fWorkspace(ws),
     fDelta(0),
     fDownX(0),
+    fStale(true),
     fDragging(false),
-    fGraphics(this),
+    fGraphics(this, true),
     fPane(d)
 {
-    addStyle(wsNoExpose);
+    addStyle(wsNoExpose | wsToolTipping);
     setParentRelative();
-    //setDND(true);
     setTitle(name());
 }
 
@@ -46,24 +58,6 @@ void WorkspaceButton::repaint() {
     fGraphics.paint();
 }
 
-void WorkspaceButton::paintBackground(Graphics& g, const YRect& r) {
-    if (taskbackPixbuf != null) {
-        g.clearArea(r.x(), r.y(), r.width(), r.height());
-        g.drawGradient(taskbackPixbuf,
-                       r.x(), r.y(), r.width(), r.height(),
-                       0, 0, width(), height());
-    }
-    else if (taskbackPixmap != null) {
-        g.fillPixmap(taskbackPixmap,
-                     r.x(), r.y(), r.width(), r.height(),
-                     r.x(), r.y());
-    }
-    else {
-        g.setColor(taskBarBg);
-        g.fillRect(r.x(), r.y(), r.width(), r.height());
-    }
-}
-
 void WorkspaceButton::handleButton(const XButtonEvent &button) {
     if (fDragging &&
         button.type == ButtonPress &&
@@ -122,6 +116,8 @@ void WorkspaceButton::handleCrossing(const XCrossingEvent &e) {
 
     if (false == pagerShowPreview) {
         super::handleCrossing(e);
+    } else {
+        YWindow::handleCrossing(e);
     }
 }
 
@@ -136,14 +132,16 @@ void WorkspaceButton::handleDNDLeave() {
     repaint();
 }
 
-void WorkspaceButton::handleBeginDrag(const XButtonEvent& d, const XMotionEvent& m)
+bool WorkspaceButton::handleBeginDrag(const XButtonEvent& d, const XMotionEvent& m)
 {
-    if (d.button == Button1) {
+    if (d.button == Button1 && fPane->limited()) {
         fDragging = true;
         fDownX = d.x_root;
         fDelta = m.x_root - d.x_root;
         fPane->drag(fWorkspace, fDelta, true, false);
+        return true;
     }
+    return false;
 }
 
 void WorkspaceButton::handleDrag(const XButtonEvent& d, const XMotionEvent& m)
@@ -180,6 +178,7 @@ void WorkspaceButton::inputReturn(YInputLine* input) {
             swap(*&str, *workspaces[fWorkspace]);
             manager->setDesktopNames();
             fPane->relabel(fWorkspace);
+            repaint();
         }
         fInput = nullptr;
     }
@@ -221,19 +220,36 @@ void WorkspaceButton::setPosition(int x, int y) {
         show();
 }
 
-WorkspacesPane::WorkspacesPane(YWindow *parent):
+WorkspacesPane::WorkspacesPane(YWindow* parent, unsigned tall):
     super(parent),
     fActive(0),
     fDelta(0),
     fMoved(0),
+    fTall(tall),
     fSpeed(0),
     fTime(zerotime()),
     fMillis(16L),
     fRepositioning(false),
-    fReconfiguring(false)
+    fReconfiguring(false),
+    fRepaintSpaces(false),
+    fDesktop(desktop->dimension()),
+    fButtons(workspaceCount)
 {
     addStyle(wsNoExpose);
     setParentRelative();
+    createButtons();
+}
+
+void WorkspacesPane::createButtons() {
+    fReconfiguring = true;
+    unsigned width = 0;
+    unsigned height = max(smallIconSize + 8, fTall);
+    for (int i = 0, n = workspaceCount; i < n; ++i) {
+        width += create(i, height)->width();
+    }
+    resize(width, height);
+    setPressed(manager->activeWorkspace(), true);
+    fReconfiguring = false;
 }
 
 void WorkspacesPane::resize(unsigned width, unsigned height) {
@@ -254,6 +270,8 @@ long WorkspacesPane::limitWidth(long paneWidth) {
     const char* str = taskBarWorkspacesLimit;
     long taskBarWidth = desktop->getScreenGeometry().width()
                       * taskBarWidthPercentage / 100;
+    long horizontally = taskBarWorkspacesLeft == leftToRight ?
+                        taskBarWidth - x() : x() + long(width());
     long reserved = 200L;
     long maxPixels = 0;
     long maxButtons = 0;
@@ -262,7 +280,7 @@ long WorkspacesPane::limitWidth(long paneWidth) {
         char* end = nullptr;
         long num = strtol(str, &end, 0);
         if (end && str < end && inrange(num, 0L, long(SHRT_MAX))) {
-            maxPixels = max(50L, taskBarWidth - x() - reserved);
+            maxPixels = max(50L, horizontally - reserved);
             maxButtons = maxPixels * count() / non_zero(paneWidth);
             maxPercent = 100L * maxPixels / taskBarWidth;
             if (*end == ' ')
@@ -282,9 +300,9 @@ long WorkspacesPane::limitWidth(long paneWidth) {
     }
     if (isEmpty(str) || 0 == (maxPixels | maxButtons)) {
         if (taskBarDoubleHeight && taskBarWorkspacesTop) {
-            maxPixels = taskBarWidth - x() - reserved;
+            maxPixels = horizontally - reserved;
         } else {
-            maxPixels = (taskBarWidth - x() - reserved) / 2;
+            maxPixels = (horizontally - reserved) / 2;
             maxButtons = 20;    // old limit
         }
     }
@@ -355,23 +373,29 @@ void WorkspacesPane::relabelButtons() {
         repositionButtons();
 
     paths = null;
+    repaint();
 }
 
 void WorkspacesPane::configure(const YRect2& r) {
     if ((fReconfiguring | fRepositioning) == false && r.resized()) {
-        fReconfiguring = true;
-        if (count() == 0) {
-            unsigned width = 0, height = max(smallIconSize + 8, r.height());
-            for (int i = 0, n = workspaceCount; i < n; ++i)
-                width += create(i, height)->width();
-            resize(width, height);
-            setPressed(manager->activeWorkspace(), true);
-            paths = null;
-        }
-        else {
+        if (pagerShowPreview && r.hh != r.old.hh && 1 < r.old.hh) {
+            rescaleButtons();
+        } else {
             repositionButtons();
         }
-        fReconfiguring = false;
+    }
+}
+
+void WorkspacesPane::rescaleButtons() {
+    for (auto wk : fButtons) {
+        scale(wk, max(height(), fTall));
+    }
+    repositionButtons();
+    for (auto wk : fButtons) {
+        if (wk->x() < int(width()))
+            wk->show();
+        else
+            break;
     }
 }
 
@@ -379,8 +403,17 @@ void WorkspacesPane::updateButtons() {
     MSG(("WorkspacesPane::udpateButtons(): updating %d -> %d",
          count(), int(workspaceCount)));
 
-    if (count() > workspaceCount)
-        fButtons.shrink(max<int>(1, workspaceCount));
+    while (workspaceCount < count() && 1 < count()) {
+        fButtons.remove(rightToLeft ? 0 : count() - 1);
+    }
+
+    if (fDesktop != desktop->dimension()) {
+        fDesktop = desktop->dimension();
+        if (pagerShowPreview) {
+            fMoved = 0;
+            rescaleButtons();
+        }
+    }
 
     int width = extent() - fMoved;
     for (int i = count(), n = workspaceCount; i < n; ++i) {
@@ -390,6 +423,7 @@ void WorkspacesPane::updateButtons() {
         if (wk->extent() > 0 && wk->x() < max(width, int(this->width())))
             wk->show();
     }
+
     resize(width, height());
     int limit = int(this->width());
     if (fMoved + width < limit) {
@@ -400,12 +434,18 @@ void WorkspacesPane::updateButtons() {
     paths = null;
 }
 
+void WorkspacesPane::scale(WorkspaceButton* button, unsigned height) {
+    unsigned dw = desktop->width();
+    unsigned dh = desktop->height();
+    unsigned scaled = (height * dw + (dh / 2)) / dh;
+    button->setSize(scaled, height);
+}
+
 WorkspaceButton* WorkspacesPane::create(int workspace, unsigned height) {
     WorkspaceButton *wk = new WorkspaceButton(workspace, this, this);
-    fButtons += wk;
+    fButtons.insert(rightToLeft ? 0 : count(), wk);
     if (pagerShowPreview) {
-        double scaled = double(height * desktop->width()) / desktop->height();
-        wk->setSize(unsigned(lround(scaled)), height);
+        scale(wk, height);
         wk->updateName();
     } else {
         label(wk);
@@ -416,9 +456,10 @@ WorkspaceButton* WorkspacesPane::create(int workspace, unsigned height) {
 }
 
 WorkspaceIcons::WorkspaceIcons() {
-    ref<YResourcePaths> dirs(YResourcePaths::subdirs("workspace", false));
-    for (int i = 0; i < dirs->getCount(); ++i) {
-        upath path(dirs->getPath(i) + "/workspace/");
+    MStringArray dirs;
+    wmapp->subdirs("workspace", false, dirs);
+    for (const mstring& base : dirs) {
+        upath path(base + "/workspace/");
         for (adir dir(path.string()); dir.next(); ) {
             files += (path + dir.entry()).string();
         }
@@ -452,17 +493,17 @@ void WorkspacesPane::label(WorkspaceButton* wk) {
 }
 
 void WorkspacesPane::relabel(int ws) {
-    label(fButtons[ws]);
+    label(index(ws));
     repositionButtons();
     paths = null;
 }
 
-void WorkspacesPane::setPressed(long ws, bool set) {
-    if (inrange<long>(1+ws, 1, count())) {
-        WorkspaceButton* wk = fButtons[int(ws)];
+void WorkspacesPane::setPressed(int ws, bool set) {
+    if (inrange(ws, 0, count() - 1)) {
+        WorkspaceButton* wk = index(ws);
         wk->setPressed(set);
         if (set) {
-            fActive = int(ws);
+            fActive = ws;
             if (wk->extent() > int(width())) {
                 fMoved -= wk->extent() - int(width());
                 repositionButtons();
@@ -475,6 +516,10 @@ void WorkspacesPane::setPressed(long ws, bool set) {
     }
 }
 
+bool WorkspacesPane::limited() const {
+    return 0 < count() && (last()->extent() - fButtons[0]->x()) > int(width());
+}
+
 void WorkspacesPane::drag(int ws, int dx, bool start, bool end) {
     int previous = fMoved;
     int farthest = extent() - fButtons[0]->x();
@@ -522,8 +567,9 @@ bool WorkspacesPane::handleTimer(YTimer *t) {
         if (fRepaintSpaces) {
             fRepaintSpaces = false;
             if (pagerShowPreview) {
-                for (IterType wk = iterator(); ++wk; ) {
-                    if (wk->visible() &&
+                for (auto wk : fButtons) {
+                    if (wk->isStale() &&
+                        wk->visible() &&
                         wk->extent() > 0 &&
                         wk->x() < int(width()))
                     {
@@ -543,18 +589,31 @@ void WorkspacesPane::stopDrag() {
     }
 }
 
-ref<YFont> WorkspaceButton::getFont() {
-    return isPressed()
-        ? (*activeWorkspaceFontName || *activeWorkspaceFontNameXft)
-        ? activeButtonFont != null
-        ? activeButtonFont
-        : activeButtonFont = YFont::getFont(XFA(activeWorkspaceFontName))
-        : YButton::getFont()
-        : (*normalWorkspaceFontName || *normalWorkspaceFontNameXft)
-        ? normalButtonFont != null
-        ? normalButtonFont
-        : normalButtonFont = YFont::getFont(XFA(normalWorkspaceFontName))
-        : YButton::getFont();
+YFont WorkspaceButton::getActiveFont() {
+    if (activeButtonFont == null) {
+        if (activeWorkspaceFontName.nonempty()) {
+            activeButtonFont = activeWorkspaceFontName;
+        }
+        if (activeButtonFont == null) {
+            activeButtonFont = YButton::getActiveFont();
+        }
+    }
+    return activeButtonFont;
+}
+
+YFont WorkspaceButton::getFont() {
+    if (isPressed()) {
+        return getActiveFont();
+    }
+    else if (normalButtonFont == null) {
+        if (normalWorkspaceFontName.nonempty()) {
+            normalButtonFont = normalWorkspaceFontName;
+        }
+        if (normalButtonFont == null) {
+            normalButtonFont = YButton::getFont();
+        }
+    }
+    return normalButtonFont;
 }
 
 YColor WorkspaceButton::getColor() {
@@ -563,14 +622,38 @@ YColor WorkspaceButton::getColor() {
         : normalButtonFg ? normalButtonFg : YButton::getColor();
 }
 
+ref<YImage> WorkspaceButton::getGradient() {
+    if (isPressed()) {
+        if (activeGradient == null ?
+            workspacebuttonactivePixbuf != null :
+            (activeGradient->width() != width() ||
+             activeGradient->height() != height()))
+            activeGradient =
+                workspacebuttonactivePixbuf->scale(width(), height());
+        return activeGradient;
+    } else {
+        if (normalGradient == null ?
+            workspacebuttonPixbuf != null :
+            (normalGradient->width() != width() ||
+             normalGradient->height() != height()))
+            normalGradient = workspacebuttonPixbuf->scale(width(), height());
+        return normalGradient;
+    }
+}
+
 YSurface WorkspaceButton::getSurface() {
     return (isPressed()
             ? YSurface(activeButtonBg ? activeButtonBg : activeBackupBg,
                                    workspacebuttonactivePixmap,
-                                   workspacebuttonactivePixbuf)
+                                   getGradient())
             : YSurface(normalButtonBg ? normalButtonBg : normalBackupBg,
                        workspacebuttonPixmap,
-                       workspacebuttonPixbuf));
+                       getGradient()));
+}
+
+YDimension WorkspaceButton::getTextSize() {
+    YFont font(getActiveFont());
+    return YDimension(font->textWidth(name()), font->height());
 }
 
 const char* WorkspaceButton::name() const {
@@ -586,19 +669,36 @@ mstring WorkspaceButton::baseName() {
     return name;
 }
 
-void WorkspaceButton::updateName() {
+void WorkspaceButton::updateToolTip() {
     setToolTip(_("Workspace: ") + baseName());
 }
 
-void WorkspacesPane::repaint() {
-    if (!pagerShowPreview) return;
+void WorkspaceButton::updateName() {
+    if (toolTipVisible()) {
+        updateToolTip();
+    }
+}
 
-    fRepaintSpaces = true;
-    fRepaintTimer->setTimer(20L, this, true);
+void WorkspacesPane::repaint() {
+    if (pagerShowPreview) {
+        fRepaintSpaces = true;
+        fRepaintTimer->setTimer(20L, this, true);
+    }
+}
+
+void WorkspacesPane::repaintWorkspace(int ws) {
+    if (pagerShowPreview) {
+        if (ws == AllWorkspaces) {
+            ws = manager->activeWorkspace();
+        }
+        if (inrange(ws, 0, count() - 1)) {
+            fButtons[ws]->setStale();
+        }
+    }
 }
 
 void WorkspaceButton::paint(Graphics &g, const YRect& r) {
-    paintBackground(g, r);
+    fStale = false;
 
     if (!pagerShowPreview) {
         YButton::paint(g, r);
@@ -616,9 +716,6 @@ void WorkspaceButton::paint(Graphics &g, const YRect& r) {
             x += 1; y += 1; w -= 2; h -= 2;
         }
 
-        unsigned wx, wy, ww, wh;
-        double sf = (double) desktop->width() / w;
-
         ref<YIcon> icon;
         YColor colors[] = {
             surface.color,
@@ -633,16 +730,23 @@ void WorkspaceButton::paint(Graphics &g, const YRect& r) {
                 yfw && yfw->getActiveLayer() <= WinLayerDock;
                 yfw = yfw->prevLayer()) {
             if (yfw->isHidden() ||
-                    !yfw->visibleOn(fWorkspace) ||
-                    hasbit(yfw->frameOptions(),
-                        YFrameWindow::foIgnoreWinList |
-                        YFrameWindow::foIgnorePagerPreview))
+                yfw->isSkipPager() ||
+                hasbit(yfw->frameOptions(),
+                       YFrameWindow::foIgnoreWinList |
+                       YFrameWindow::foIgnorePagerPreview)) {
                 continue;
-            wx = (unsigned) round(double(yfw->x()) / sf) + x;
-            wy = (unsigned) round(double(yfw->y()) / sf) + y;
-            ww = (unsigned) round(double(yfw->width()) / sf);
-            wh = (unsigned) round(double(yfw->height()) / sf);
-            if (ww < 1 || wh < 1)
+            }
+            if (yfw->isAllWorkspaces() ?
+                fWorkspace != manager->activeWorkspace() :
+                fWorkspace != yfw->getWorkspace()) {
+                continue;
+            }
+            int dw = int(desktop->width());
+            int wx = x + (yfw->x() * w + (dw / 2)) / dw;
+            int wy = y + (yfw->y() * w + (dw / 2)) / dw;
+            int ww = (int(yfw->width()) * w + (dw / 2)) / dw;
+            int wh = (int(yfw->height()) * w + (dw / 2)) / dw;
+            if (ww <= 1 || wh <= 1)
                 continue;
             if (yfw->isMaximizedVert()) { // !!! hack
                 wy = y; wh = h;
@@ -659,9 +763,12 @@ void WorkspaceButton::paint(Graphics &g, const YRect& r) {
                         g.setColor(colors[2]);
                     g.fillRect(wx+1, wy+1, ww-2, wh-2);
 
-                    if (pagerShowWindowIcons && ww > smallIconSize+1 &&
-                            wh > smallIconSize+1 && (icon = yfw->clientIcon()) != null &&
-                            icon->small() != null) {
+                    if (pagerShowWindowIcons &&
+                        ww > 1 + int(smallIconSize) &&
+                        wh > 1 + int(smallIconSize) &&
+                        (icon = yfw->getIcon()) != null &&
+                        icon->small() != null)
+                    {
                         g.drawImage(icon->small(),
                                     wx + (ww-smallIconSize)/2,
                                     wy + (wh-smallIconSize)/2);
@@ -688,16 +795,20 @@ void WorkspaceButton::paint(Graphics &g, const YRect& r) {
             strlcpy(label, baseName(), min(5, int(sizeof label)));
         }
         if (label[0] != 0) {
-            ref<YFont> font = getFont();
+            YFont font = getFont();
+            if (font) {
+                int wx = (w - font->textWidth(label)) / 2 + x;
+                int wy = (h - font->height()) / 2 + font->ascent() + y;
 
-            wx = (w - font->textWidth(label)) / 2 + x;
-            wy = (h - font->height()) / 2 + font->ascent() + y;
-
-            g.setFont(font);
-            g.setColor(colors[0]);
-            g.drawChars(label, 0, strlen(label), wx+1, wy+1);
-            g.setColor(colors[3]);
-            g.drawChars(label, 0, strlen(label), wx, wy);
+                g.setFont(font);
+                g.setColor(colors[0]);
+                g.drawString(wx+1, wy+1, label);
+                g.setColor(colors[3]);
+                g.drawString(wx, wy, label);
+            }
+        }
+        if (surface.gradient != null) {
+            g.maxOpacity();
         }
     }
 }
