@@ -3,8 +3,8 @@
 
 #include "WinMgr.h"
 #include "ylist.h"
-#include "yaction.h"
 #include "ymsgbox.h"
+#include "ypopup.h"
 #include "workspaces.h"
 
 extern YAction layerActionSet[WinLayerCount];
@@ -14,59 +14,25 @@ class YWindowManager;
 class YFrameClient;
 class YFrameWindow;
 class YSMListener;
+class SwitchWindow;
+class MiniIcon;
+class DockApp;
 class IApp;
 
-/*
- * X11 time state to support _NET_WM_USER_TIME.
- * Keep track of the time in seconds when we receive a X11 time stamp.
- * Only compare two X11 time stamps if they are in a time interval.
- */
-class UserTime {
-private:
-    unsigned long xtime;
-    bool valid;
-    long since;
-    enum {
-        XTimeMask = 0xFFFFFFFFUL,       // milliseconds
-        XTimeRange = 0x7FFFFFFFUL,      // milliseconds
-        SInterval = 0x3FFFFFFFUL / 1000,     // seconds
-    };
-public:
-    UserTime() : xtime(0), valid(false), since(0) { }
-    explicit UserTime(unsigned long xtime, bool valid = true) :
-        xtime(xtime & XTimeMask), valid(valid), since(seconds()) { }
-    unsigned long time() const { return xtime; }
-    bool good() const { return valid; }
-    long when() const { return since; }
-    bool update(unsigned long xtime, bool valid = true) {
-        UserTime u(xtime, valid);
-        return *this < u || xtime == 0 ? (*this = u, true) : false;
-    }
-    bool operator<(const UserTime& u) const {
-        if (since == 0 || u.since == 0) return u.since != 0;
-        if (valid == false || u.valid == false) return u.valid;
-        if (since < u.since && u.since - since > SInterval) return true;
-        if (since > u.since && since - u.since > SInterval) return false;
-        if (xtime < u.xtime) return u.xtime - xtime <= XTimeRange;
-        if (xtime > u.xtime) return xtime - u.xtime >  XTimeRange;
-        return false;
-    }
-    bool operator==(const UserTime& u) const {
-        return !(*this < u) && !(u < *this);
-    }
-};
-
-class EdgeSwitch: public YWindow, public YTimerListener {
+class EdgeSwitch: public YDndWindow, public YTimerListener {
 public:
     EdgeSwitch(YWindowManager *manager, int delta, bool vertical);
     virtual ~EdgeSwitch();
 
     virtual void handleCrossing(const XCrossingEvent &crossing);
     virtual bool handleTimer(YTimer *t);
+    virtual void handleDNDEnter();
+    virtual void handleDNDLeave();
     void setGeometry();
+    int destination();
 private:
     YWindowManager *fManager;
-    YCursor & fCursor;
+    Cursor fCursor;
     int fDelta;
     bool fVert;
 
@@ -81,7 +47,37 @@ public:
     virtual void handleButton(const XButtonEvent &button);
 };
 
-class YWindowManager: public YDesktop, private YMsgBoxListener {
+class YArrange {
+public:
+    YArrange(YFrameWindow** w, int n) : win(w), count(n) { }
+    void discard() { delete[] win; win = nullptr; count = 0; }
+    int size() const { return count; }
+    YFrameWindow** begin() const { return win; }
+    YFrameWindow** end() const { return win + count; }
+    YFrameWindow* operator[](int index) const { return win[index]; }
+    operator bool() const { return win && count; }
+private:
+    YFrameWindow** win;
+    int count;
+};
+
+class YTopWindow : public YWindow {
+public:
+    YTopWindow();
+    bool handleKey(const XKeyEvent& key) override;
+    void setFrame(YFrameWindow* frame);
+
+private:
+    YFrameWindow* fFrame;
+    Window fHandle;
+};
+
+class YWindowManager:
+    private YDesktop,
+    private YMsgBoxListener,
+    private YTimerListener,
+    public YPopDownListener
+{
 public:
     YWindowManager(
         IApp *app,
@@ -101,6 +97,7 @@ public:
     virtual void handleConfigure(const XConfigureEvent &configure);
     virtual void handleConfigureRequest(const XConfigureRequestEvent &configureRequest);
     virtual void handleMapRequest(const XMapRequestEvent &mapRequest);
+    virtual void handleMapNotify(const XMapEvent& map);
     virtual void handleUnmapNotify(const XUnmapEvent &unmap);
     virtual void handleClientMessage(const XClientMessageEvent &message);
     virtual void handleProperty(const XPropertyEvent &property);
@@ -110,6 +107,8 @@ public:
     virtual void handleRRNotify(const XRRNotifyEvent &notify);
 #endif
     virtual void handleMsgBox(YMsgBox *msgbox, int operation);
+    virtual void handlePopDown(YPopupWindow* popup);
+    virtual bool handleTimer(YTimer* timer);
 
     void manageClients();
     void unmanageClients();
@@ -122,14 +121,15 @@ public:
 
     YFrameWindow *findFrame(Window win);
     YFrameClient *findClient(Window win);
-    YFrameWindow *manageClient(Window win, bool mapClient = false);
+    void manageClient(YFrameClient* client, bool mapClient = false);
     void unmanageClient(YFrameClient *client);
-    void destroyedClient(Window win);
-    YFrameWindow *mapClient(Window win);
+    void clientDestroyed(YFrameClient* client);
+    void clientTransfered(YFrameClient* client, YFrameWindow* frame);
+    void mapClient(Window win);
 
-    void setFocus(YFrameWindow *f, bool canWarp = false);
-    YFrameWindow *getFocus() { return fFocusWin; }
-    void loseFocus(YFrameWindow *window);
+    void setFocus(YFrameWindow *f, bool canWarp = false, bool reorder = true);
+    YFrameWindow* getFocus() const { return fFocusWin; }
+    ClientData* getFocused() const;
 
     void installColormap(Colormap cmap);
     void setColormapWindow(YFrameWindow *frame);
@@ -138,26 +138,28 @@ public:
     void removeClientFrame(YFrameWindow *frame);
 
     void updateScreenSize(XEvent *event);
-    void getWorkArea(YFrameWindow *frame, int *mx, int *my, int *Mx, int *My, int xiscreen = -1);
+    void getWorkArea(int *mx, int *my, int *Mx, int *My);
+    void getWorkArea(const YFrameWindow *frame, int *mx, int *my, int *Mx, int *My, int xiscreen = -1);
     void getWorkAreaSize(YFrameWindow *frame, int *Mw,int *Mh);
 
     int calcCoverage(bool down, YFrameWindow *frame, int x, int y, int w, int h);
     void tryCover(bool down, YFrameWindow *frame, int x, int y, int w, int h,
-                  int &px, int &py, int &cover, int xiscreen);
+                  int& px, int& py, int& cover, int mx, int my, int Mx, int My);
     bool getSmartPlace(bool down, YFrameWindow *frame, int &x, int &y, int w, int h, int xiscreen);
     void getNewPosition(YFrameWindow *frame, int &x, int &y, int w, int h, int xiscreen);
     void placeWindow(YFrameWindow *frame, int x, int y, int cw, int ch, bool newClient, bool &canActivate);
 
-    YFrameWindow *top(long layer) const;
-    void setTop(long layer, YFrameWindow *top);
-    YFrameWindow *bottom(long layer) const;
-    void setBottom(long layer, YFrameWindow *bottom);
+    YFrameWindow* top(int layer) const;
+    void setTop(int layer, YFrameWindow *top);
+    YFrameWindow* bottom(int layer) const;
+    void setBottom(int layer, YFrameWindow *bottom);
+    YWindow* bottomWindow() const { return fBottom; }
 
-    YFrameWindow *topLayer(long layer = WinLayerCount - 1);
-    YFrameWindow *bottomLayer(long layer = 0);
+    YFrameWindow* topLayer(int layer = WinLayerCount - 1);
+    YFrameWindow* bottomLayer(int layer = 0);
 
-    void setAbove(YFrameWindow* frame, YFrameWindow* above);
-    void setBelow(YFrameWindow* frame, YFrameWindow* below);
+    bool setAbove(YFrameWindow* frame, YFrameWindow* above);
+    bool setBelow(YFrameWindow* frame, YFrameWindow* below);
     void removeLayeredFrame(YFrameWindow *);
     void appendCreatedFrame(YFrameWindow *f);
     void removeCreatedFrame(YFrameWindow *f);
@@ -170,28 +172,25 @@ public:
     void lowerFocusFrame(YFrameWindow* frame);
     void raiseFocusFrame(YFrameWindow* frame);
 
-    void restackWindows(YFrameWindow *win);
+    void restackWindows();
     void focusTopWindow();
-    YFrameWindow *getFrameUnderMouse(long workspace = AllWorkspaces);
-    YFrameWindow *getLastFocus(bool skipAllWorkspaces = false, long workspace = AllWorkspaces);
+    YFrameWindow *getFrameUnderMouse(int workspace = AllWorkspaces);
+    YFrameWindow *getLastFocus(bool skipAllWorkspaces = false,
+                               int workspace = AllWorkspaces);
     void focusLastWindow();
     bool focusTop(YFrameWindow *f);
-    void relocateWindows(long workspace, int screen, int dx, int dy);
     void updateClientList();
     void updateUserTime(const UserTime& userTime);
-
-    YMenu *createWindowMenu(YMenu *menu, long workspace);
-    int windowCount(long workspace);
     void popupWindowListMenu(YWindow *owner, int x, int y);
 
     void initWorkspaces();
 
-    long activeWorkspace() const { return fActiveWorkspace; }
-    long lastWorkspace() const { return fLastWorkspace; }
-    void activateWorkspace(long workspace);
+    int activeWorkspace() const { return fActiveWorkspace; }
+    int lastWorkspace() const { return fLastWorkspace; }
+    void activateWorkspace(int workspace);
 
-    void appendNewWorkspaces(long extra);
-    void removeLastWorkspaces(long minus);
+    void extendWorkspaces(int target);
+    void lessenWorkspaces(int target);
     void updateWorkspaces(bool increase);
 
     void setShowingDesktop();
@@ -200,31 +199,30 @@ public:
     void updateTaskBarNames();
     void updateMoveMenu();
 
-    bool readCurrentDesktop(long &workspace);
+    bool readCurrentDesktop(int &workspace);
     void setDesktopGeometry();
     bool compareDesktopNames(const YStringList& list);
     bool readDesktopLayout();
     void readDesktopNames(bool init, bool net);
     bool readNetDesktopNames(YStringList& list);
     bool readWinDesktopNames(YStringList& list);
-    void setWinDesktopNames(long count);
-    void setNetDesktopNames(long count);
-    void setDesktopNames(long count);
+    void setNetDesktopNames(int count);
+    void setDesktopNames(int count);
     void setDesktopNames();
     void setDesktopCount();
     void setDesktopViewport();
 
     void announceWorkArea();
-    void setWinWorkspace(int workspace);
     void updateWorkArea();
-    void updateWorkAreaInner();
+    bool updateWorkAreaInner();
     void debugWorkArea(const char* prefix);
+    void workAreaUpdated();
     void resizeWindows();
 
-    void getIconPosition(YFrameWindow *frame, int *iconX, int *iconY);
+    void getIconPosition(MiniIcon* iw, int *iconX, int *iconY);
 
     void wmCloseSession();
-    void exitAfterLastClient(bool shuttingDown);
+    void exitAfterLastClient(bool exitWhenDone);
     void checkLogout();
 
     virtual void resetColormap(bool active);
@@ -236,36 +234,39 @@ public:
     void popupStartMenu(YWindow *owner);
     void popupWindowListMenu(YWindow *owner);
 
-    void switchToWorkspace(long nw, bool takeCurrent);
+    void switchToWorkspace(int nw, bool takeCurrent);
     void switchToPrevWorkspace(bool takeCurrent);
     void switchToNextWorkspace(bool takeCurrent);
     void switchToLastWorkspace(bool takeCurrent);
 
     void tilePlace(YFrameWindow *w, int tx, int ty, int tw, int th);
-    void tileWindows(YFrameWindow **w, int count, bool vertical);
-    void smartPlace(YFrameWindow **w, int count);
+    void tileWindows(YArrange arrange, bool vertical);
+    void smartPlace(YArrange arrange);
     void getCascadePlace(YFrameWindow *frame, int &lastX, int &lastY, int &x, int &y, int w, int h);
-    void cascadePlace(YFrameWindow **w, int count);
-    void setWindows(YFrameWindow **w, int count, YAction action);
+    void cascadePlace(YArrange arrange);
+    void setWindows(YArrange arrange, YAction action);
 
-    bool getWindowsToArrange(YFrameWindow ***w, int *count, bool sticky = false, bool skipNonMinimizable = false);
+    YArrange getWindowsToArrange(bool sticky = false, bool mini = false, bool full = false);
 
-    void saveArrange(YFrameWindow **w, int count);
+    bool saveArrange(YArrange arrange);
     void undoArrange();
+    void arrangeIcons();
+    void tileWindows(bool vertical);
+    void arrangeWindows();
+    void actionWindows(YAction action);
+    void toggleDesktop();
+    void cascadeWindows();
 
     bool haveClients();
     void setupRootProxy();
-    void setWorkAreaMoveWindows(bool m) { fWorkAreaMoveWindows = m; }
     void setKeyboard(mstring keyboard);
     void setKeyboard(int configIndex);
     mstring getKeyboard();
     void updateKeyboard(int configIndex);
+    void reflectKeyboard(int configIndex, mstring keyboard);
+    void kbLayout();
 
-    void updateFullscreenLayer();
-    void updateFullscreenLayerEnable(bool enable);
-    int getSwitchScreen();
-
-    static void doWMAction(WMAction action);
+    static void doWMAction(WMAction action, bool putback = false);
     void lockFocus() {
         //MSG(("lockFocus %d", lockFocusCount));
         lockFocusCount++;
@@ -280,10 +281,24 @@ public:
         if (0 == --fWorkAreaLock && fWorkAreaUpdate)
             updateWorkArea();
     }
+    void requestWorkAreaUpdate() { ++fWorkAreaUpdate; }
+
+    void lockRestack() { fRestackLock++; }
+    void unlockRestack() {
+        if (0 == --fRestackLock && fRestackUpdate) {
+            fRestackUpdate = 0;
+            restackWindows();
+        }
+    }
 
     enum WMState { wmSTARTUP, wmRUNNING, wmSHUTDOWN };
 
     WMState wmState() const { return fWmState; }
+    bool isStartup() const { return fWmState == wmSTARTUP; }
+    bool isRunning() const { return fWmState == wmRUNNING; }
+    bool notRunning() const { return fWmState != wmRUNNING; }
+    bool notShutting() const { return fWmState != wmSHUTDOWN; }
+    bool shuttingDown() const { return fWmState == wmSHUTDOWN; }
     bool fullscreenEnabled() { return fFullscreenEnabled; }
     void setFullscreenEnabled(bool enable) { fFullscreenEnabled = enable; }
     const UserTime& lastUserTime() const { return fLastUserTime; }
@@ -298,6 +313,13 @@ public:
     };
 
     const DesktopLayout& layout() const { return fLayout; }
+    bool handleSwitchWorkspaceKey(const XKeyEvent& key, KeySym k, unsigned vm);
+
+    int getSwitchScreen();
+    bool switchWindowVisible() const;
+    SwitchWindow* getSwitchWindow();
+    Window netActiveWindow() const { return fActiveWindow; }
+    int edgeWorkspace(int x, int y);
 
 private:
     struct WindowPosState {
@@ -306,10 +328,15 @@ private:
         YFrameWindow *frame;
     };
 
-    void updateArea(long workspace, int screen_number, int l, int t, int r, int b);
-    bool handleWMKey(const XKeyEvent &key, KeySym k, unsigned int m, unsigned int vm);
+    bool ignoreOverride(Window win, const XWindowAttributes& attr, int* layer);
+    bool tabbingClient(YFrameClient* client);
+    YFrameClient* allocateClient(Window win, bool mapClient);
+    YFrameWindow* allocateFrame(YFrameClient* client);
+    void updateArea(int workspace, int screen_number, int l, int t, int r, int b);
+    bool handleWMKey(const XKeyEvent &key, KeySym k, unsigned vm);
     void setWmState(WMState newWmState);
     void refresh();
+    void focusOverlap();
 
     IApp *app;
     YActionListener *wmActionListener;
@@ -320,11 +347,11 @@ private:
     YCreatedList fCreationOrder;  // frame creation order
     YFocusedList fFocusedOrder;   // focus order: old -> now
 
-    long fActiveWorkspace;
-    long fLastWorkspace;
+    int fActiveWorkspace;
+    int fLastWorkspace;
     YFrameWindow *fColormapWindow;
 
-    long fWorkAreaWorkspaceCount;
+    int fWorkAreaWorkspaceCount;
     int fWorkAreaScreenCount;
     struct WorkAreaRect {
         int fMinX, fMinY, fMaxX, fMaxY;
@@ -350,16 +377,21 @@ private:
     } **fWorkArea;
 
     YObjectArray<EdgeSwitch> edges;
-    bool fShuttingDown;
+    bool fExitWhenDone;
     int fArrangeCount;
     WindowPosState *fArrangeInfo;
     YProxyWindow *rootProxy;
-    YWindow *fTopWin;
-    bool fWorkAreaMoveWindows;
-    bool fOtherScreenFocused;
+    YTopWindow *fTopWin;
+    YWindow *fBottom;
+    int fCascadeX;
+    int fCascadeY;
+    int fIconColumn;
+    int fIconRow;
     int lockFocusCount;
     int fWorkAreaLock;
     int fWorkAreaUpdate;
+    int fRestackLock;
+    int fRestackUpdate;
     int fServerGrabCount;
     bool fFullscreenEnabled;
 
@@ -372,29 +404,32 @@ private:
     DesktopLayout fLayout;
     mstring fCurrentKeyboard;
     int fDefaultKeyboard;
+    SwitchWindow* fSwitchWindow;
+    lazy<YTimer> fSwitchDownTimer;
+    lazy<YTimer> fLayoutTimer;
+    DockApp* fDockApp;
 };
 
 extern YWindowManager *manager;
 
+class YRestackLock {
+public:
+    YRestackLock() { manager->lockRestack(); }
+    ~YRestackLock() { manager->unlockRestack(); }
+};
+
+class YFullscreenLock {
+public:
+    YFullscreenLock() { manager->setFullscreenEnabled(false); }
+    ~YFullscreenLock() { manager->setFullscreenEnabled(true); }
+};
+
 void dumpZorder(const char *oper, YFrameWindow *w, YFrameWindow *a = nullptr);
 
-extern Atom _XA_WIN_APP_STATE;
-extern Atom _XA_WIN_AREA_COUNT;
-extern Atom _XA_WIN_AREA;
-extern Atom _XA_WIN_CLIENT_LIST;
-extern Atom _XA_WIN_DESKTOP_BUTTON_PROXY;
-extern Atom _XA_WIN_EXPANDED_SIZE;
-extern Atom _XA_WIN_HINTS;
 extern Atom _XA_WIN_ICONS;
 extern Atom _XA_WIN_LAYER;
 extern Atom _XA_WIN_PROTOCOLS;
-extern Atom _XA_WIN_STATE;
-extern Atom _XA_WIN_SUPPORTING_WM_CHECK;
 extern Atom _XA_WIN_TRAY;
-extern Atom _XA_WIN_WORKAREA;
-extern Atom _XA_WIN_WORKSPACE_COUNT;
-extern Atom _XA_WIN_WORKSPACE_NAMES;
-extern Atom _XA_WIN_WORKSPACE;
 
 extern Atom _XA_WM_CLIENT_LEADER;
 extern Atom _XA_WM_CLIENT_MACHINE;
@@ -404,6 +439,7 @@ extern Atom _XA_ICEWM_ACTION;
 extern Atom _XA_ICEWM_GUIEVENT;
 extern Atom _XA_ICEWM_HINT;
 extern Atom _XA_ICEWM_FONT_PATH;
+extern Atom _XA_ICEWM_TABS;
 extern Atom _XA_XROOTPMAP_ID;
 extern Atom _XA_XROOTCOLOR_PIXEL;
 
